from .azure import AzureRealtimeBetaLLMService
from .events import (
    InputAudioNoiseReduction,
    InputAudioTranscription,
    SemanticTurnDetection,
    SessionProperties,
    TurnDetection,
)
from .openai import OpenAIRealtimeBetaLLMService
